# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Windows SDK functions for Microsoft Visual C/C++.
"""

import os

from ..common import (
    debug,
)

from . import Util
from . import Config
from . import Registry

from .Exceptions import (
    MSVCInternalError,
)

from . import Dispatcher
Dispatcher.register_modulename(__name__)


_DESKTOP = Config.MSVC_PLATFORM_INTERNAL['Desktop']
_UWP = Config.MSVC_PLATFORM_INTERNAL['UWP']

def _new_sdk_map():
    sdk_map = {
        _DESKTOP.vc_platform: [],
        _UWP.vc_platform: [],
    }
    return sdk_map

def _sdk_10_layout(version):

    folder_prefix = version + '.'

    sdk_map = _new_sdk_map()

    sdk_roots = Registry.sdk_query_paths(version)

    sdk_version_platform_seen = set()
    sdk_roots_seen = set()

    for sdk_t in sdk_roots:

        sdk_root = sdk_t[0]
        if sdk_root in sdk_roots_seen:
            continue
        sdk_roots_seen.add(sdk_root)

        if not os.path.exists(sdk_root):
            continue

        sdk_include_path = os.path.join(sdk_root, 'include')
        if not os.path.exists(sdk_include_path):
            continue

        for version_nbr, version_nbr_path in Util.listdir_dirs(sdk_include_path):

            if not version_nbr.startswith(folder_prefix):
                continue

            sdk_inc_path = Util.normalize_path(os.path.join(version_nbr_path, 'um'))
            if not os.path.exists(sdk_inc_path):
                continue

            for vc_platform, sdk_inc_file in [
                (_DESKTOP.vc_platform, 'winsdkver.h'),
                (_UWP.vc_platform,     'windows.h'),
            ]:

                if not os.path.exists(os.path.join(sdk_inc_path, sdk_inc_file)):
                    continue

                key = (version_nbr, vc_platform)
                if key in sdk_version_platform_seen:
                    continue
                sdk_version_platform_seen.add(key)

                sdk_map[vc_platform].append(version_nbr)

    for key, val in sdk_map.items():
        val.sort(reverse=True)

    return sdk_map

def _sdk_81_layout(version):

    version_nbr = version

    sdk_map = _new_sdk_map()

    sdk_roots = Registry.sdk_query_paths(version)

    sdk_version_platform_seen = set()
    sdk_roots_seen = set()

    for sdk_t in sdk_roots:

        sdk_root = sdk_t[0]
        if sdk_root in sdk_roots_seen:
            continue
        sdk_roots_seen.add(sdk_root)

        # msvc does not check for existence of root or other files

        sdk_inc_path = Util.normalize_path(os.path.join(sdk_root, r'include\um'))
        if not os.path.exists(sdk_inc_path):
            continue

        for vc_platform, sdk_inc_file in [
            (_DESKTOP.vc_platform, 'winsdkver.h'),
            (_UWP.vc_platform,     'windows.h'),
        ]:

            if not os.path.exists(os.path.join(sdk_inc_path, sdk_inc_file)):
                continue

            key = (version_nbr, vc_platform)
            if key in sdk_version_platform_seen:
                continue
            sdk_version_platform_seen.add(key)

            sdk_map[vc_platform].append(version_nbr)

    for key, val in sdk_map.items():
        val.sort(reverse=True)

    return sdk_map

_sdk_map_cache = {}
_sdk_cache = {}

def _reset_sdk_cache():
    global _sdk_map_cache
    global _sdk_cache
    debug('')
    _sdk_map_cache = {}
    _sdk_cache = {}

def _sdk_10(key, reg_version):
    if key in _sdk_map_cache:
        sdk_map = _sdk_map_cache[key]
    else:
        sdk_map = _sdk_10_layout(reg_version)
        _sdk_map_cache[key] = sdk_map
    return sdk_map

def _sdk_81(key, reg_version):
    if key in _sdk_map_cache:
        sdk_map = _sdk_map_cache[key]
    else:
        sdk_map = _sdk_81_layout(reg_version)
        _sdk_map_cache[key] = sdk_map
    return sdk_map

def _combine_sdk_map_list(sdk_map_list):
    combined_sdk_map = _new_sdk_map()
    for sdk_map in sdk_map_list:
        for key, val in sdk_map.items():
            combined_sdk_map[key].extend(val)
    return combined_sdk_map

_sdk_dispatch_map = {
    '10.0': (_sdk_10, '10.0'),
    '8.1': (_sdk_81, '8.1'),
}

def _verify_sdk_dispatch_map():
    debug('')
    for sdk_version in Config.MSVC_SDK_VERSIONS:
        if sdk_version in _sdk_dispatch_map:
            continue
        err_msg = f'sdk version {sdk_version} not in sdk_dispatch_map'
        raise MSVCInternalError(err_msg)
    return None

def _version_list_sdk_map(version_list):
    sdk_map_list = []
    for version in version_list:
        func, reg_version = _sdk_dispatch_map[version]
        sdk_map = func(version, reg_version)
        sdk_map_list.append(sdk_map)

    combined_sdk_map = _combine_sdk_map_list(sdk_map_list)
    return combined_sdk_map

def _sdk_map(version_list):
    key = tuple(version_list)
    if key in _sdk_cache:
        sdk_map = _sdk_cache[key]
    else:
        version_numlist = [float(v) for v in version_list]
        version_numlist.sort(reverse=True)
        key = tuple([str(v) for v in version_numlist])
        sdk_map = _version_list_sdk_map(key)
        _sdk_cache[key] = sdk_map
    return sdk_map

def get_msvc_platform(is_uwp=False):
    platform_def = _UWP if is_uwp else _DESKTOP
    return platform_def

def get_sdk_version_list(vs_def, platform_def):
    version_list = vs_def.vc_sdk_versions if vs_def.vc_sdk_versions is not None else []
    sdk_map = _sdk_map(version_list)
    sdk_list = sdk_map.get(platform_def.vc_platform, [])
    return sdk_list

def get_msvc_sdk_version_list(msvc_version, msvc_uwp_app=False):
    debug('msvc_version=%s, msvc_uwp_app=%s', repr(msvc_version), repr(msvc_uwp_app))

    sdk_versions = []

    verstr = Util.get_msvc_version_prefix(msvc_version)
    if not verstr:
        debug('msvc_version is not defined')
        return sdk_versions

    vs_def = Config.MSVC_VERSION_EXTERNAL.get(verstr, None)
    if not vs_def:
        debug('vs_def is not defined')
        return sdk_versions

    is_uwp = True if msvc_uwp_app in Config.BOOLEAN_SYMBOLS[True] else False
    platform_def = get_msvc_platform(is_uwp)
    sdk_list = get_sdk_version_list(vs_def, platform_def)

    sdk_versions.extend(sdk_list)
    debug('sdk_versions=%s', repr(sdk_versions))

    return sdk_versions

def reset():
    debug('')
    _reset_sdk_cache()

def verify():
    debug('')
    _verify_sdk_dispatch_map()




#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""GitHub interfacing for nuitka-watch."""

import os
import sys

from nuitka.tools.quality.Git import getModifiedPaths, getRemoteURL
from nuitka.Tracing import tools_logger
from nuitka.utils.Execution import callProcess, check_call


def checkInNuitkaWatch():
    remote_url = getRemoteURL("origin")
    assert remote_url in (
        "git@github.com:Nuitka/Nuitka-Watch.git",
        "https://github.com/Nuitka/Nuitka-Watch",
    ), remote_url
    assert os.path.exists(".git")


def createNuitkaWatchPR(category, description):
    checkInNuitkaWatch()

    modified_files = list(getModifiedPaths())

    if not modified_files:
        tools_logger.sysexit("Nothing to do", exit_code=0)

    changed_flavors = set()

    for modified_file in modified_files:
        if os.path.basename(modified_file) == "compilation-report.xml":
            flavor = os.path.basename(os.path.dirname(modified_file))
            changed_flavors.add(flavor)

    if not changed_flavors:
        tools_logger.sysexit("No changes in compilation reports, only other things.")

    if len(changed_flavors) != 1:
        tools_logger.sysexit("Only a single flavor is supported at a time currently.")

    (changed_flavor,) = changed_flavors

    tools_logger.info(
        "Detected changes for %s in results of '%s'." % (description, changed_flavor)
    )

    commit_message = """
Changes for %s in results of '%s'

This change is automatically generated and the result of executing %s on
current Nuitka-Watch state.
""" % (
        description,
        changed_flavor,
        " ".join(sys.argv),
    )

    branch_name = "auto-%s-%s" % (category, changed_flavor)

    # May not exist of course.
    callProcess(["git", "branch", "-D", branch_name])

    check_call(["git", "branch", branch_name])

    check_call(["git", "checkout", branch_name])

    check_call(["git", "add", "."])

    check_call(["git", "commit", "-m", commit_message])

    check_call(["git", "checkout", "main"])

    tools_logger.info("Change is now on branch '%s'." % branch_name)


if __name__ == "__main__":
    # TODO: This runner should be directly used from nuitka-watch binary,
    # but for development purposes, we keep it separately accessible.

    from nuitka.options.CommandLineOptionsTools import makeOptionsParser

    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--desc",
        action="store",
        dest="desc",
        help="""\
Description of the change, e.g. "Nuitka update 1.9.3".""",
    )

    options, positional_args = parser.parse_args()
    assert not positional_args

    createNuitkaWatchPR(category="hotfix", description=options.desc)

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Stage changes in Nuitka-Watch automatically

This aims at recognizing unimportant changes automatically and is used
for larger report format migrations in the future potentially.
"""

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.quality.Git import (
    getCheckoutFileChangeDesc,
    getFileHashContent,
    putFileHashContent,
    updateGitFile,
)
from nuitka.TreeXML import convertStringToXML, convertXmlToString
from nuitka.utils.FileOperations import getFileContents, withTemporaryFile

options = None


# This is XML API, spell-checker: ignore addprevious,getnext,addnext


def _findMatchingChild(current, node):
    node_tag = node.tag

    if node_tag == "module":
        attrib_name = "name"
    elif node_tag == "optimization-time":
        attrib_name = "pass"
    elif node_tag == "code-generation-time":
        attrib_name = None
    else:
        assert False, (node, current)

    if attrib_name is not None:
        attrib_value = node.attrib[attrib_name]

    for candidate in current.findall(node_tag):
        if attrib_name is not None:
            try:
                candidate_value = candidate.attrib[attrib_name]
            except KeyError:
                assert False, current

            if candidate_value == attrib_value:
                return candidate
        else:
            return candidate

    return None


def findMatchingNode(root, search_node):
    """Find a node matching the given one in a XML root."""
    nodes = []

    node = search_node
    while node is not None:
        nodes.insert(0, node)

        node = node.getparent()

    # Root is easy and hard coded.
    current = root
    del nodes[0]

    for node in nodes:
        current = _findMatchingChild(current, node)

        if current is None:
            return None

    return current


def _acceptOptimizationTimeChanges(old_report, new_report):
    changed = False

    for new_node in new_report.xpath("//module/optimization-time"):
        old_node = findMatchingNode(old_report, new_node)

        if old_node is not None:
            old_node.getparent().replace(old_node, new_node)
            changed = True

    for new_node in new_report.xpath("//module/code-generation-time"):
        old_node = findMatchingNode(old_report, new_node)

        if old_node is not None:
            old_node.getparent().replace(old_node, new_node)
            changed = True
        else:
            parent_module = findMatchingNode(old_report, new_node.getparent())

            if parent_module is not None:
                optimization_times = parent_module.findall("optimization-time")

                if optimization_times:
                    parent_module.insert(
                        parent_module.index(optimization_times[-1]) + 1, new_node
                    )
                else:
                    parent_module.append(new_node)

                changed = True

    return changed


def _acceptDllOrderChanges(old_report, new_report):
    changed = False

    processed_parents = set()
    for new_dll in new_report.xpath("//included_dll"):
        new_parent = new_dll.getparent()
        if new_parent in processed_parents:
            continue
        processed_parents.add(new_parent)

        old_parent = findMatchingNode(old_report, new_parent)

        if old_parent is None:
            continue

        last_node = None
        used_old_nodes = set()

        for new_node in new_parent.findall("included_dll"):
            old_node = None
            for candidate in old_parent.findall("included_dll"):
                if candidate in used_old_nodes:
                    continue

                if candidate.attrib == new_node.attrib:
                    old_node = candidate
                    used_old_nodes.add(old_node)
                    break

            if old_node is not None:
                if last_node is None:
                    first_existing = old_parent.find("included_dll")
                    if first_existing is not None and first_existing != old_node:
                        first_existing.addprevious(old_node)
                        changed = True
                else:
                    if last_node.getnext() != old_node:
                        last_node.addnext(old_node)
                        changed = True

                last_node = old_node

    return changed


def onCompilationReportChange(filename, git_stage):
    print("Working on", filename)

    new_report = convertStringToXML(getFileContents(filename, mode="rb"), use_lxml=True)
    old_git_contents = getFileHashContent(git_stage["src_hash"])

    old_report = convertStringToXML(old_git_contents, use_lxml=True)

    new_nuitka_version = new_report.attrib["nuitka_version"]
    changed = False
    if old_report.attrib["nuitka_version"] != new_nuitka_version:
        old_report.attrib["nuitka_version"] = new_nuitka_version

        changed = True

    if options.accept_optimization_time:
        changed = changed | _acceptOptimizationTimeChanges(old_report, new_report)

    if options.accept_dll_order:
        changed = changed | _acceptDllOrderChanges(old_report, new_report)

    if changed:
        new_git_contents = convertXmlToString(old_report)
        with withTemporaryFile(mode="w", delete=False) as output_file:
            tmp_filename = output_file.name
            output_file.write(new_git_contents)
            output_file.close()

        new_hash_value = putFileHashContent(tmp_filename)

        if git_stage["src_hash"] != new_hash_value:
            updateGitFile(filename, git_stage["src_hash"], new_hash_value, staged=False)


def onFileChange(git_stage):
    filename = git_stage["src_path"]

    if filename.endswith("compilation-report.xml"):
        onCompilationReportChange(filename=filename, git_stage=git_stage)


def main():
    # Cheating in this singleton and not passing options,
    # pylint: disable=global-statement
    global options

    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--accept-optimization-time",
        action="store_true",
        dest="accept_optimization_time",
        default=False,
        help="""Accept module optimization-time and code-generation-time changes.""",
    )

    # TODO: Once the sorting in reports is done, remove this option again.
    parser.add_option(
        "--accept-dll-order",
        action="store_true",
        dest="accept_dll_order",
        default=False,
        help="""Accept DLL order changes.""",
    )

    options, _positional_args = parser.parse_args()

    for git_stage in getCheckoutFileChangeDesc(staged=False):
        onFileChange(git_stage=git_stage)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Generic Taskmaster module for the SCons build engine.

This module contains the primary interface(s) between a wrapping user
interface and the SCons build engine.  There are two key classes here:

Taskmaster
    This is the main engine for walking the dependency graph and
    calling things to decide what does or doesn't need to be built.

Task
    This is the base class for allowing a wrapping interface to
    decide what does or doesn't actually need to be done.  The
    intention is for a wrapping interface to subclass this as
    appropriate for different types of behavior it may need.

    The canonical example is the SCons native Python interface,
    which has Task subclasses that handle its specific behavior,
    like printing "'foo' is up to date" when a top-level target
    doesn't need to be built, and handling the -c option by removing
    targets as its "build" action.  There is also a separate subclass
    for suppressing this output when the -q option is used.

    The Taskmaster instantiates a Task object for each (set of)
    target(s) that it decides need to be evaluated and/or built.
"""
import io
import sys
from abc import ABC, abstractmethod
from itertools import chain
import logging

import SCons.Errors
import SCons.Node
import SCons.Warnings
from SCons.Util import DispatchingFormatter

StateString = SCons.Node.StateString
NODE_NO_STATE = SCons.Node.no_state
NODE_PENDING = SCons.Node.pending
NODE_EXECUTING = SCons.Node.executing
NODE_UP_TO_DATE = SCons.Node.up_to_date
NODE_EXECUTED = SCons.Node.executed
NODE_FAILED = SCons.Node.failed
print_prepare = False               # set by option --debug=prepare

# A subsystem for recording stats about how different Nodes are handled by
# the main Taskmaster loop.  There's no external control here (no need for
# a --debug= option); enable it by changing the value of CollectStats.

CollectStats = None


class Stats:
    """
    A simple class for holding statistics about the disposition of a
    Node by the Taskmaster.  If we're collecting statistics, each Node
    processed by the Taskmaster gets one of these attached, in which case
    the Taskmaster records its decision each time it processes the Node.
    (Ideally, that's just once per Node.)
    """
    def __init__(self):
        """
        Instantiates a Taskmaster.Stats object, initializing all
        appropriate counters to zero.
        """
        self.considered  = 0
        self.already_handled  = 0
        self.problem  = 0
        self.child_failed  = 0
        self.not_built  = 0
        self.side_effects  = 0
        self.build  = 0


StatsNodes = []

fmt = "%(considered)3d "\
      "%(already_handled)3d " \
      "%(problem)3d " \
      "%(child_failed)3d " \
      "%(not_built)3d " \
      "%(side_effects)3d " \
      "%(build)3d "


def dump_stats():
    for n in sorted(StatsNodes, key=lambda a: str(a)):
        print((fmt % n.attributes.stats.__dict__) + str(n))


class Task(ABC):
    """ SCons build engine abstract task class.

    This controls the interaction of the actual building of node
    and the rest of the engine.

    This is expected to handle all of the normally-customizable
    aspects of controlling a build, so any given application
    *should* be able to do what it wants by sub-classing this
    class and overriding methods as appropriate.  If an application
    needs to customize something by sub-classing Taskmaster (or
    some other build engine class), we should first try to migrate
    that functionality into this class.

    Note that it's generally a good idea for sub-classes to call
    these methods explicitly to update state, etc., rather than
    roll their own interaction with Taskmaster from scratch.
    """

    LOGGER = None

    def __init__(self, tm, targets, top, node):
        self.tm = tm
        self.targets = targets
        self.top = top
        self.node = node
        self.exc_clear()

    def trace_message(self, node, description='node'):
        # This grabs the name of the function which calls trace_message()
        method_name=sys._getframe(1).f_code.co_name+"():"
        Task.LOGGER.debug('%-15s %s %s' % (method_name, description, self.tm.tm_trace_node(node)))

    def display(self, message):
        """
        Hook to allow the calling interface to display a message.

        This hook gets called as part of preparing a task for execution
        (that is, a Node to be built).  As part of figuring out what Node
        should be built next, the actual target list may be altered,
        along with a message describing the alteration.  The calling
        interface can subclass Task and provide a concrete implementation
        of this method to see those messages.
        """
        pass

    def prepare(self):
        """
        Called just before the task is executed.

        This is mainly intended to give the target Nodes a chance to
        unlink underlying files and make all necessary directories before
        the Action is actually called to build the targets.
        """
        global print_prepare
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        # Now that it's the appropriate time, give the TaskMaster a
        # chance to raise any exceptions it encountered while preparing
        # this task.
        self.exception_raise()

        if self.tm.message:
            self.display(self.tm.message)
            self.tm.message = None

        # Let the targets take care of any necessary preparations.
        # This includes verifying that all of the necessary sources
        # and dependencies exist, removing the target file(s), etc.
        #
        # As of April 2008, the get_executor().prepare() method makes
        # sure that all of the aggregate sources necessary to build this
        # Task's target(s) exist in one up-front check.  The individual
        # target t.prepare() methods check that each target's explicit
        # or implicit dependencies exists, and also initialize the
        # .sconsign info.
        executor = self.targets[0].get_executor()
        if executor is None:
            return
        executor.prepare()
        for t in executor.get_action_targets():
            if print_prepare:
                print("Preparing target %s..."%t)
                for s in t.side_effects:
                    print("...with side-effect %s..."%s)
            t.prepare()
            for s in t.side_effects:
                if print_prepare:
                    print("...Preparing side-effect %s..."%s)
                s.prepare()

    def get_target(self):
        """Fetch the target being built or updated by this task.
        """
        return self.node

    @abstractmethod
    def needs_execute(self):
        return

    def execute(self):
        """
        Called to execute the task.

        This method is called from multiple threads in a parallel build,
        so only do thread safe stuff here.  Do thread unsafe stuff in
        prepare(), executed() or failed().
        """
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        try:
            cached_targets = []
            for t in self.targets:
                if not t.retrieve_from_cache():
                    break
                cached_targets.append(t)
            if len(cached_targets) < len(self.targets):
                # Remove targets before building. It's possible that we
                # partially retrieved targets from the cache, leaving
                # them in read-only mode. That might cause the command
                # to fail.
                #
                for t in cached_targets:
                    try:
                        t.fs.unlink(t.get_internal_path())
                    except OSError as e:
                        SCons.Warnings.warn(SCons.Warnings.CacheCleanupErrorWarning,
                            "Failed copying all target files from cache, Error while attempting to remove file %s retrieved from cache: %s" % (t.get_internal_path(), e))
                self.targets[0].build()
                for t in self.targets:
                    t.push_to_cache()
            else:
                for t in cached_targets:
                    t.cached = 1
        except SystemExit:
            exc_value = sys.exc_info()[1]
            raise SCons.Errors.ExplicitExit(self.targets[0], exc_value.code)
        except SCons.Errors.UserError:
            raise
        except SCons.Errors.BuildError:
            raise
        except Exception as e:
            # Nuitka: This is only bug hiding, e.g. in case of clcache.
            raise

            buildError = SCons.Errors.convert_to_BuildError(e)
            buildError.node = self.targets[0]
            buildError.exc_info = sys.exc_info()
            raise buildError

    def executed_without_callbacks(self):
        """
        Called when the task has been successfully executed
        and the Taskmaster instance doesn't want to call
        the Node's callback methods.
        """
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        for t in self.targets:
            if t.get_state() == NODE_EXECUTING:
                for side_effect in t.side_effects:
                    side_effect.set_state(NODE_NO_STATE)
                t.set_state(NODE_EXECUTED)

    def executed_with_callbacks(self):
        """
        Called when the task has been successfully executed and
        the Taskmaster instance wants to call the Node's callback
        methods.

        This may have been a do-nothing operation (to preserve build
        order), so we must check the node's state before deciding whether
        it was "built", in which case we call the appropriate Node method.
        In any event, we always call "visited()", which will handle any
        post-visit actions that must take place regardless of whether
        or not the target was an actual built target or a source Node.
        """
        global print_prepare
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        for t in self.targets:
            if t.get_state() == NODE_EXECUTING:
                for side_effect in t.side_effects:
                    side_effect.set_state(NODE_NO_STATE)
                t.set_state(NODE_EXECUTED)
                t.built()
                t.visited()
                if (not print_prepare and
                    (not hasattr(self, 'options') or not self.options.debug_includes)):
                    t.release_target_info()
            else:
                t.visited()

    executed = executed_with_callbacks

    def failed(self):
        """
        Default action when a task fails:  stop the build.

        Note: Although this function is normally invoked on nodes in
        the executing state, it might also be invoked on up-to-date
        nodes when using Configure().
        """
        self.fail_stop()

    def fail_stop(self):
        """
        Explicit stop-the-build failure.

        This sets failure status on the target nodes and all of
        their dependent parent nodes.

        Note: Although this function is normally invoked on nodes in
        the executing state, it might also be invoked on up-to-date
        nodes when using Configure().
        """
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        # Invoke will_not_build() to clean-up the pending children
        # list.
        self.tm.will_not_build(self.targets, lambda n: n.set_state(NODE_FAILED))

        # Tell the taskmaster to not start any new tasks
        self.tm.stop()

        # We're stopping because of a build failure, but give the
        # calling Task class a chance to postprocess() the top-level
        # target under which the build failure occurred.
        self.targets = [self.tm.current_top]
        self.top = 1

    def fail_continue(self):
        """
        Explicit continue-the-build failure.

        This sets failure status on the target nodes and all of
        their dependent parent nodes.

        Note: Although this function is normally invoked on nodes in
        the executing state, it might also be invoked on up-to-date
        nodes when using Configure().
        """
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        self.tm.will_not_build(self.targets, lambda n: n.set_state(NODE_FAILED))

    def make_ready_all(self):
        """
        Marks all targets in a task ready for execution.

        This is used when the interface needs every target Node to be
        visited--the canonical example being the "scons -c" option.
        """
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        self.out_of_date = self.targets[:]
        for t in self.targets:
            t.disambiguate().set_state(NODE_EXECUTING)
            for s in t.side_effects:
                # add disambiguate here to mirror the call on targets above
                s.disambiguate().set_state(NODE_EXECUTING)

    def make_ready_current(self):
        """
        Marks all targets in a task ready for execution if any target
        is not current.

        This is the default behavior for building only what's necessary.
        """
        global print_prepare
        T = self.tm.trace
        if T:
            T.log_handler.stream.write('\n') # Prefix message with new line. This is a hack
            self.trace_message(self.node)

        self.out_of_date = []
        needs_executing = False
        for t in self.targets:
            try:
                t.disambiguate().make_ready()
                is_up_to_date = not t.has_builder() or \
                                (not t.always_build and t.is_up_to_date())
            except OSError as e:
                raise SCons.Errors.BuildError(node=t, errstr=e.strerror, filename=e.filename)

            if not is_up_to_date:
                self.out_of_date.append(t)
                needs_executing = True

        if needs_executing:
            for t in self.targets:
                t.set_state(NODE_EXECUTING)
                for s in t.side_effects:
                    # add disambiguate here to mirror the call on targets in first loop above
                    s.disambiguate().set_state(NODE_EXECUTING)
        else:
            for t in self.targets:
                # We must invoke visited() to ensure that the node
                # information has been computed before allowing the
                # parent nodes to execute. (That could occur in a
                # parallel build...)
                t.visited()
                t.set_state(NODE_UP_TO_DATE)
                if (not print_prepare and
                    (not hasattr(self, 'options') or not self.options.debug_includes)):
                    t.release_target_info()

    make_ready = make_ready_current

    def postprocess(self):
        """
        Post-processes a task after it's been executed.

        This examines all the targets just built (or not, we don't care
        if the build was successful, or even if there was no build
        because everything was up-to-date) to see if they have any
        waiting parent Nodes, or Nodes waiting on a common side effect,
        that can be put back on the candidates list.
        """
        T = self.tm.trace
        if T:
            self.trace_message(self.node)

        # We may have built multiple targets, some of which may have
        # common parents waiting for this build.  Count up how many
        # targets each parent was waiting for so we can subtract the
        # values later, and so we *don't* put waiting side-effect Nodes
        # back on the candidates list if the Node is also a waiting
        # parent.

        targets = set(self.targets)

        pending_children = self.tm.pending_children
        parents = {}
        for t in targets:
            # A node can only be in the pending_children set if it has
            # some waiting_parents.
            if t.waiting_parents:
                if T:
                    self.trace_message(t, 'removing')
                pending_children.discard(t)
            for p in t.waiting_parents:
                parents[p] = parents.get(p, 0) + 1
            t.waiting_parents = set()

        for t in targets:
            if t.side_effects is not None:
                for s in t.side_effects:
                    if s.get_state() == NODE_EXECUTING:
                        s.set_state(NODE_NO_STATE)

                    # The side-effects may have been transferred to
                    # NODE_NO_STATE by executed_with{,out}_callbacks, but was
                    # not taken out of the waiting parents/pending children
                    # data structures. Check for that now.
                    if s.get_state() == NODE_NO_STATE and s.waiting_parents:
                        pending_children.discard(s)
                        for p in s.waiting_parents:
                            parents[p] = parents.get(p, 0) + 1
                        s.waiting_parents = set()
                    for p in s.waiting_s_e:
                        if p.ref_count == 0:
                            self.tm.candidates.append(p)

        for p, subtract in parents.items():
            p.ref_count = p.ref_count - subtract
            if T:
                self.trace_message(p, 'adjusted parent ref count')
            if p.ref_count == 0:
                self.tm.candidates.append(p)

        for t in targets:
            t.postprocess()

    # Exception handling subsystem.
    #
    # Exceptions that occur while walking the DAG or examining Nodes
    # must be raised, but must be raised at an appropriate time and in
    # a controlled manner so we can, if necessary, recover gracefully,
    # possibly write out signature information for Nodes we've updated,
    # etc.  This is done by having the Taskmaster tell us about the
    # exception, and letting

    def exc_info(self):
        """
        Returns info about a recorded exception.
        """
        return self.exception

    def exc_clear(self):
        """
        Clears any recorded exception.

        This also changes the "exception_raise" attribute to point
        to the appropriate do-nothing method.
        """
        self.exception = (None, None, None)
        self.exception_raise = self._no_exception_to_raise

    def exception_set(self, exception=None):
        """
        Records an exception to be raised at the appropriate time.

        This also changes the "exception_raise" attribute to point
        to the method that will, in fact
        """
        if not exception:
            exception = sys.exc_info()
        self.exception = exception
        self.exception_raise = self._exception_raise

    def _no_exception_to_raise(self):
        pass

    def _exception_raise(self):
        """
        Raises a pending exception that was recorded while getting a
        Task ready for execution.
        """
        exc = self.exc_info()[:]
        try:
            exc_type, exc_value, exc_traceback = exc
        except ValueError:
            exc_type, exc_value = exc  # pylint: disable=unbalanced-tuple-unpacking
            exc_traceback = None

        # raise exc_type(exc_value).with_traceback(exc_traceback)
        if isinstance(exc_value, Exception): #hasattr(exc_value, 'with_traceback'):
            # If exc_value is an exception, then just reraise
            raise exc_value.with_traceback(exc_traceback)
        else:
            # else we'll create an exception using the value and raise that
            raise exc_type(exc_value).with_traceback(exc_traceback)


        # raise e.__class__, e.__class__(e), sys.exc_info()[2]
        #     exec("raise exc_type(exc_value).with_traceback(exc_traceback)")



class AlwaysTask(Task):
    def needs_execute(self):
        """
        Always returns True (indicating this Task should always
        be executed).

        Subclasses that need this behavior (as opposed to the default
        of only executing Nodes that are out of date w.r.t. their
        dependencies) can use this as follows:

            class MyTaskSubclass(SCons.Taskmaster.Task):
                needs_execute = SCons.Taskmaster.AlwaysTask.needs_execute
        """
        return True

class OutOfDateTask(Task):
    def needs_execute(self):
        """
        Returns True (indicating this Task should be executed) if this
        Task's target state indicates it needs executing, which has
        already been determined by an earlier up-to-date check.
        """
        return self.targets[0].get_state() == SCons.Node.executing


def find_cycle(stack, visited):
    if stack[-1] in visited:
        return None
    visited.add(stack[-1])
    for n in stack[-1].waiting_parents:
        stack.append(n)
        if stack[0] == stack[-1]:
            return stack
        if find_cycle(stack, visited):
            return stack
        stack.pop()
    return None


class Taskmaster:
    """
    The Taskmaster for walking the dependency DAG.
    """

    def __init__(self, targets=[], tasker=None, order=None, trace=None):
        self.original_top = targets
        self.top_targets_left = targets[:]
        self.top_targets_left.reverse()
        self.candidates = []
        if tasker is None:
            tasker = OutOfDateTask
        self.tasker = tasker
        if not order:
            order = lambda l: l
        self.order = order
        self.message = None
        self.next_candidate = self.find_next_candidate
        self.pending_children = set()
        self.trace = False
        self.configure_trace(trace)

    def configure_trace(self, trace=None):
        """
        This handles the command line option --taskmastertrace=
        It can be:
        -           : output to stdout
        <filename>  : output to a file
        False/None  : Do not trace
        """
        if not trace:
            self.trace = False
            return

        # TODO: May want to switch format to something like this.
        # log_format = (
        #     '%(relativeCreated)05dms'
        #     ':%(relfilename)s'
        #     ':%(funcName)s'
        #     '#%(lineno)s'
        #     ': %(message)s'
        # )
        tm_formatter = logging.Formatter('Taskmaster: %(message)s')
        if isinstance(trace, io.StringIO):
            log_handler = logging.StreamHandler(trace)
        elif trace == '-':
            log_handler = logging.StreamHandler(sys.stdout)
        elif trace:
            log_handler = logging.FileHandler(filename=trace)

        logger = logging.getLogger('Taskmaster')
        logger.setLevel(level=logging.DEBUG)
        logger.addHandler(log_handler)
        self.trace = logger

        logger.log_handler = log_handler

        # Now setup Task's logger.
        tl = logging.getLogger("Task")
        tl.setLevel(level=logging.DEBUG)
        tl.addHandler(log_handler)
        task_formatter = logging.Formatter('%(name)s.%(message)s')
        Task.LOGGER = tl

        self.trace.log_handler = log_handler

        log_handler.setFormatter(DispatchingFormatter(
            formatters={
                'Taskmaster': tm_formatter,
                'Task': task_formatter,
                'Job': task_formatter,
            },
            default_formatter=logging.Formatter('%(message)s')
        ))

    def find_next_candidate(self):
        """
        Returns the next candidate Node for (potential) evaluation.

        The candidate list (really a stack) initially consists of all of
        the top-level (command line) targets provided when the Taskmaster
        was initialized.  While we walk the DAG, visiting Nodes, all the
        children that haven't finished processing get pushed on to the
        candidate list.  Each child can then be popped and examined in
        turn for whether *their* children are all up-to-date, in which
        case a Task will be created for their actual evaluation and
        potential building.

        Here is where we also allow candidate Nodes to alter the list of
        Nodes that should be examined.  This is used, for example, when
        invoking SCons in a source directory.  A source directory Node can
        return its corresponding build directory Node, essentially saying,
        "Hey, you really need to build this thing over here instead."
        """
        try:
            return self.candidates.pop()
        except IndexError:
            pass
        try:
            node = self.top_targets_left.pop()
        except IndexError:
            return None
        self.current_top = node
        alt, message = node.alter_targets()
        if alt:
            self.message = message
            self.candidates.append(node)
            self.candidates.extend(self.order(alt))
            node = self.candidates.pop()
        return node

    def no_next_candidate(self):
        """
        Stops Taskmaster processing by not returning a next candidate.

        Note that we have to clean-up the Taskmaster candidate list
        because the cycle detection depends on the fact all nodes have
        been processed somehow.
        """
        while self.candidates:
            candidates = self.candidates
            self.candidates = []
            self.will_not_build(candidates)
        return None

    def _validate_pending_children(self):
        """
        Validate the content of the pending_children set. Assert if an
        internal error is found.

        This function is used strictly for debugging the taskmaster by
        checking that no invariants are violated. It is not used in
        normal operation.

        The pending_children set is used to detect cycles in the
        dependency graph. We call a "pending child" a child that is
        found in the "pending" state when checking the dependencies of
        its parent node.

        A pending child can occur when the Taskmaster completes a loop
        through a cycle. For example, let's imagine a graph made of
        three nodes (A, B and C) making a cycle. The evaluation starts
        at node A. The Taskmaster first considers whether node A's
        child B is up-to-date. Then, recursively, node B needs to
        check whether node C is up-to-date. This leaves us with a
        dependency graph looking like::

                                          Next candidate \
                                                          \
            Node A (Pending) --> Node B(Pending) --> Node C (NoState)
                    ^                                     |
                    |                                     |
                    +-------------------------------------+

        Now, when the Taskmaster examines the Node C's child Node A,
        it finds that Node A is in the "pending" state. Therefore,
        Node A is a pending child of node C.

        Pending children indicate that the Taskmaster has potentially
        loop back through a cycle. We say potentially because it could
        also occur when a DAG is evaluated in parallel. For example,
        consider the following graph::

            Node A (Pending) --> Node B(Pending) --> Node C (Pending) --> ...
                    |                                     ^
                    |                                     |
                    +----------> Node D (NoState) --------+
                                      /
                      Next candidate /

        The Taskmaster first evaluates the nodes A, B, and C and
        starts building some children of node C. Assuming, that the
        maximum parallel level has not been reached, the Taskmaster
        will examine Node D. It will find that Node C is a pending
        child of Node D.

        In summary, evaluating a graph with a cycle will always
        involve a pending child at one point. A pending child might
        indicate either a cycle or a diamond-shaped DAG. Only a
        fraction of the nodes ends-up being a "pending child" of
        another node. This keeps the pending_children set small in
        practice.

        We can differentiate between the two cases if we wait until
        the end of the build. At this point, all the pending children
        nodes due to a diamond-shaped DAG will have been properly
        built (or will have failed to build). But, the pending
        children involved in a cycle will still be in the pending
        state.

        The taskmaster removes nodes from the pending_children set as
        soon as a pending_children node moves out of the pending
        state. This also helps to keep the pending_children set small.
        """

        for n in self.pending_children:
            assert n.state in (NODE_PENDING, NODE_EXECUTING), \
                (str(n), StateString[n.state])
            assert len(n.waiting_parents) != 0, (str(n), len(n.waiting_parents))
            for p in n.waiting_parents:
                assert p.ref_count > 0, (str(n), str(p), p.ref_count)

    def tm_trace_node(self, node):
        return('<%-10s %-3s %s>' % (StateString[node.get_state()],
                                    node.ref_count,
                                    repr(str(node))))

    def _find_next_ready_node(self):
        """
        Finds the next node that is ready to be built.

        This is *the* main guts of the DAG walk.  We loop through the
        list of candidates, looking for something that has no un-built
        children (i.e., that is a leaf Node or has dependencies that are
        all leaf Nodes or up-to-date).  Candidate Nodes are re-scanned
        (both the target Node itself and its sources, which are always
        scanned in the context of a given target) to discover implicit
        dependencies.  A Node that must wait for some children to be
        built will be put back on the candidates list after the children
        have finished building.  A Node that has been put back on the
        candidates list in this way may have itself (or its sources)
        re-scanned, in order to handle generated header files (e.g.) and
        the implicit dependencies therein.

        Note that this method does not do any signature calculation or
        up-to-date check itself.  All of that is handled by the Task
        class.  This is purely concerned with the dependency graph walk.
        """

        self.ready_exc = None

        T = self.trace
        if T:
            T.log_handler.stream.write('\n')  # Prefix message with new line. This is a hack
            self.trace.debug('Looking for a node to evaluate')

        while True:
            node = self.next_candidate()
            if node is None:
                if T:
                    self.trace.debug('No candidate anymore.')
                return None

            node = node.disambiguate()
            state = node.get_state()

            # For debugging only:
            #
            # try:
            #     self._validate_pending_children()
            # except:
            #     self.ready_exc = sys.exc_info()
            #     return node

            if CollectStats:
                if not hasattr(node.attributes, 'stats'):
                    node.attributes.stats = Stats()
                    StatsNodes.append(node)
                S = node.attributes.stats
                S.considered = S.considered + 1
            else:
                S = None

            if T:
                self.trace.debug('    Considering node %s and its children:' % self.tm_trace_node(node))

            if state == NODE_NO_STATE:
                # Mark this node as being on the execution stack:
                node.set_state(NODE_PENDING)
            elif state > NODE_PENDING:
                # Skip this node if it has already been evaluated:
                if S: S.already_handled = S.already_handled + 1
                if T:
                    self.trace.debug('       already handled (executed)')
                continue

            executor = node.get_executor()

            try:
                children = executor.get_all_children()
            except SystemExit:
                exc_value = sys.exc_info()[1]
                e = SCons.Errors.ExplicitExit(node, exc_value.code)
                self.ready_exc = (SCons.Errors.ExplicitExit, e)
                if T:
                    self.trace.debug('       SystemExit')
                return node
            except Exception as e:
                # We had a problem just trying to figure out the
                # children (like a child couldn't be linked in to a
                # VariantDir, or a Scanner threw something).  Arrange to
                # raise the exception when the Task is "executed."
                self.ready_exc = sys.exc_info()
                if S: S.problem = S.problem + 1
                if T:
                    self.trace.debug('       exception %s while scanning children.' % e)
                return node

            children_not_visited = []
            children_pending = set()
            children_not_ready = []
            children_failed = False

            for child in chain(executor.get_all_prerequisites(), children):
                childstate = child.get_state()

                if T:
                    self.trace.debug('       ' + self.tm_trace_node(child))

                if childstate == NODE_NO_STATE:
                    children_not_visited.append(child)
                elif childstate == NODE_PENDING:
                    children_pending.add(child)
                elif childstate == NODE_FAILED:
                    children_failed = True

                if childstate <= NODE_EXECUTING:
                    children_not_ready.append(child)

            # These nodes have not even been visited yet.  Add
            # them to the list so that on some next pass we can
            # take a stab at evaluating them (or their children).
            if children_not_visited:
                if len(children_not_visited) > 1:
                    children_not_visited.reverse()
                self.candidates.extend(self.order(children_not_visited))

            # if T and children_not_visited:
            #    self.trace.debug('     adding to candidates: %s' % map(str, children_not_visited))
            #    self.trace.debug('     candidates now: %s\n' % map(str, self.candidates))

            # Skip this node if any of its children have failed.
            #
            # This catches the case where we're descending a top-level
            # target and one of our children failed while trying to be
            # built by a *previous* descent of an earlier top-level
            # target.
            #
            # It can also occur if a node is reused in multiple
            # targets. One first descends though the one of the
            # target, the next time occurs through the other target.
            #
            # Note that we can only have failed_children if the
            # --keep-going flag was used, because without it the build
            # will stop before diving in the other branch.
            #
            # Note that even if one of the children fails, we still
            # added the other children to the list of candidate nodes
            # to keep on building (--keep-going).
            if children_failed:
                for n in executor.get_action_targets():
                    n.set_state(NODE_FAILED)

                if S: S.child_failed = S.child_failed + 1
                if T:
                    self.trace.debug('****** %s' % self.tm_trace_node(node))
                continue

            if children_not_ready:
                for child in children_not_ready:
                    # We're waiting on one or more derived targets
                    # that have not yet finished building.
                    if S: S.not_built = S.not_built + 1

                    # Add this node to the waiting parents lists of
                    # anything we're waiting on, with a reference
                    # count so we can be put back on the list for
                    # re-evaluation when they've all finished.
                    node.ref_count =  node.ref_count + child.add_to_waiting_parents(node)
                    if T:
                        self.trace.debug('     adjusted ref count: %s, child %s' %
                                                (self.tm_trace_node(node), repr(str(child))))

                if T:
                    for pc in children_pending:
                        self.trace.debug('       adding %s to the pending children set' %
                                              self.tm_trace_node(pc))
                self.pending_children = self.pending_children | children_pending

                continue

            # Skip this node if it has side-effects that are
            # currently being built:
            wait_side_effects = False
            for se in executor.get_action_side_effects():
                if se.get_state() == NODE_EXECUTING:
                    se.add_to_waiting_s_e(node)
                    wait_side_effects = True

            if wait_side_effects:
                if S: S.side_effects = S.side_effects + 1
                continue

            # The default when we've gotten through all of the checks above:
            # this node is ready to be built.
            if S: S.build = S.build + 1
            if T:
                self.trace.debug('Evaluating %s' % self.tm_trace_node(node))

            # For debugging only:
            #
            # try:
            #     self._validate_pending_children()
            # except:
            #     self.ready_exc = sys.exc_info()
            #     return node

            return node

        return None

    def next_task(self):
        """
        Returns the next task to be executed.

        This simply asks for the next Node to be evaluated, and then wraps
        it in the specific Task subclass with which we were initialized.
        """
        node = self._find_next_ready_node()

        if node is None:
            return None

        executor = node.get_executor()
        if executor is None:
            return None

        tlist = executor.get_all_targets()

        task = self.tasker(self, tlist, node in self.original_top, node)
        try:
            task.make_ready()
        except Exception as e :
            # We had a problem just trying to get this task ready (like
            # a child couldn't be linked to a VariantDir when deciding
            # whether this node is current).  Arrange to raise the
            # exception when the Task is "executed."
            self.ready_exc = sys.exc_info()

        if self.ready_exc:
            task.exception_set(self.ready_exc)

        self.ready_exc = None

        return task

    def will_not_build(self, nodes, node_func=lambda n: None):
        """
        Perform clean-up about nodes that will never be built. Invokes
        a user defined function on all of these nodes (including all
        of their parents).
        """

        T = self.trace

        pending_children = self.pending_children

        to_visit = set(nodes)
        pending_children = pending_children - to_visit

        if T:
            for n in nodes:
                self.trace.debug('       removing node %s from the pending children set\n' %
                                      self.tm_trace_node(n))
        try:
            while len(to_visit):
                node = to_visit.pop()
                node_func(node)

                # Prune recursion by flushing the waiting children
                # list immediately.
                parents = node.waiting_parents
                node.waiting_parents = set()

                to_visit = to_visit | parents
                pending_children = pending_children - parents

                for p in parents:
                    p.ref_count = p.ref_count - 1
                    if T:
                        self.trace.debug('       removing parent %s from the pending children set\n' %
                                                self.tm_trace_node(p))
        except KeyError:
            # The container to_visit has been emptied.
            pass

        # We have the stick back the pending_children list into the
        # taskmaster because the python 1.5.2 compatibility does not
        # allow us to use in-place updates
        self.pending_children = pending_children

    def stop(self):
        """
        Stops the current build completely.
        """
        self.next_candidate = self.no_next_candidate

    def cleanup(self):
        """
        Check for dependency cycles.
        """
        if not self.pending_children:
            return

        nclist = [(n, find_cycle([n], set())) for n in self.pending_children]

        genuine_cycles = [
            node for node,cycle in nclist
                     if cycle or node.get_state() != NODE_EXECUTED
        ]
        if not genuine_cycles:
            # All of the "cycles" found were single nodes in EXECUTED state,
            # which is to say, they really weren't cycles.  Just return.
            return

        desc = 'Found dependency cycle(s):\n'
        for node, cycle in nclist:
            if cycle:
                desc = desc + "  " + " -> ".join(map(str, cycle)) + "\n"
            else:
                desc = desc + \
                    "  Internal Error: no cycle found for node %s (%s) in state %s\n" %  \
                    (node, repr(node), StateString[node.get_state()])

        raise SCons.Errors.UserError(desc)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Write and read constants data files and provide identifiers."""

import pickle
import sys

from nuitka.__past__ import (  # pylint: disable=I0021,redefined-builtin
    BaseExceptionGroup,
    ExceptionGroup,
    GenericAlias,
    UnionType,
    basestring,
    to_byte,
    xrange,
)
from nuitka.Builtins import (
    builtin_anon_codes,
    builtin_anon_values,
    builtin_exception_values_list,
)

# TODO: Move to constants
from nuitka.code_generation.GlobalConstants import getConstantDefaultPopulation
from nuitka.code_generation.Namify import namifyConstant
from nuitka.containers.OrderedSets import OrderedSet
from nuitka.OutputDirectories import getSourceDirectoryPath
from nuitka.PythonVersions import python_version
from nuitka.utils.FileOperations import getNormalizedPathJoin, openPickleFile


class BuiltinAnonValue(object):
    """Used to pickle anonymous values."""

    # Python3 values has no index, turn into a tuple.
    anon_values = tuple(builtin_anon_values.values())

    def __init__(self, anon_name):
        self.anon_name = anon_name

    def getStreamValueByte(self):
        """Return byte value, encoding the anon built-in value."""

        return to_byte(self.anon_values.index(self.anon_name))


class BuiltinGenericAliasValue(object):
    """For transporting GenericAlias values through pickler."""

    def __init__(self, origin, args):
        self.origin = origin
        self.args = args


class BuiltinUnionTypeValue(object):
    """For transporting UnionType values through pickler."""

    def __init__(self, args):
        self.args = args


class BuiltinSpecialValue(object):
    """Used to pickle special values."""

    def __init__(self, value):
        self.value = value

    def getStreamValueByte(self):
        """Return byte value, encoding the special built-in value."""

        # Currently the only ones.
        if self.value == "Ellipsis":
            return to_byte(0)
        elif self.value == "NotImplemented":
            return to_byte(1)
        elif self.value == "Py_SysVersionInfo":
            return to_byte(2)
        else:
            assert False, self.value


class BlobData(object):
    """Used to pickle bytes to become raw pointers."""

    __slots__ = ("data", "name")

    def __init__(self, data, name):
        self.data = data
        self.name = name

    def getData(self):
        return self.data

    def __repr__(self):
        return "<nuitka.Serialization.BlobData %s>" % self.name


def _pickleAnonValues(pickler, value):
    if value in builtin_anon_values:
        pickler.save(BuiltinAnonValue(builtin_anon_values[value]))
    elif value is Ellipsis:
        pickler.save(BuiltinSpecialValue("Ellipsis"))
    elif value is NotImplemented:
        pickler.save(BuiltinSpecialValue("NotImplemented"))
    elif value is sys.version_info:
        pickler.save(BuiltinSpecialValue("Py_SysVersionInfo"))
    else:
        pickler.save_global(value)


def _pickleGenericAlias(pickler, value):
    pickler.save(BuiltinGenericAliasValue(origin=value.__origin__, args=value.__args__))


def _pickleUnionType(pickler, value):
    pickler.save(BuiltinUnionTypeValue(args=value.__args__))


class ConstantStreamWriter(object):
    """Write constants to a stream and return numbers for them."""

    def __init__(self, filename):
        self.count = 0

        filename = getNormalizedPathJoin(
            getSourceDirectoryPath(onefile=False, create=False),
            filename,
        )

        self.file, self.pickle = openPickleFile(filename, "wb")

        self.pickle.dispatch[type] = _pickleAnonValues
        self.pickle.dispatch[type(Ellipsis)] = _pickleAnonValues
        self.pickle.dispatch[type(NotImplemented)] = _pickleAnonValues

        if type(sys.version_info) is not tuple:
            self.pickle.dispatch[type(sys.version_info)] = _pickleAnonValues

        # Standard pickling doesn't work with our necessary wrappers.
        if python_version >= 0x390:
            self.pickle.dispatch[GenericAlias] = _pickleGenericAlias

        if python_version >= 0x3A0:
            self.pickle.dispatch[UnionType] = _pickleUnionType

    def addConstantValue(self, constant_value):
        self.pickle.dump(constant_value)
        self.count += 1

    def addBlobData(self, data, name):
        self.pickle.dump(BlobData(data, name))
        self.count += 1

    def close(self):
        self.file.close()


class ConstantStreamReader(object):
    def __init__(self, const_file):
        # spell-checker: ignore Unpickler

        self.count = 0
        self.pickle = pickle.Unpickler(const_file)

    def readConstantValue(self):
        return self.pickle.load()


class GlobalConstantAccessor(object):
    __slots__ = ("constants", "constants_writer", "top_level_name")

    global_constant_keys = set()

    def __init__(self, data_filename, top_level_name):
        self.constants = OrderedSet()

        self.constants_writer = ConstantStreamWriter(data_filename)
        self.top_level_name = top_level_name

        # We need to delay this, as getConstantDefaultPopulation() cannot be
        # called at import time for this module, wants to know e.g. deployment
        # mode, and other options.
        if not GlobalConstantAccessor.global_constant_keys:
            GlobalConstantAccessor.global_constant_keys.update(
                "const_" + namifyConstant(constant)
                for constant in getConstantDefaultPopulation()
            )

    def getConstantCode(self, constant):
        # Use in user code, or for constants building code itself, many constant
        # types get special code immediately, and it's return driven.
        # pylint: disable=too-many-branches,too-many-return-statements
        if constant is None:
            return "Py_None"
        elif constant is True:
            return "Py_True"
        elif constant is False:
            return "Py_False"
        elif constant is Ellipsis:
            return "Py_Ellipsis"
        elif constant is NotImplemented:
            return "Py_NotImplemented"
        elif constant is sys.version_info:
            return "Py_SysVersionInfo"
        elif type(constant) is type:
            # TODO: Maybe make this a mapping in nuitka.Builtins

            if constant is None:
                return "(PyObject *)Py_TYPE(Py_None)"
            elif constant is object:
                return "(PyObject *)&PyBaseObject_Type"
            elif constant is staticmethod:
                return "(PyObject *)&PyStaticMethod_Type"
            elif constant is classmethod:
                return "(PyObject *)&PyClassMethod_Type"
            elif constant is bytearray:
                return "(PyObject *)&PyByteArray_Type"
            elif constant is enumerate:
                return "(PyObject *)&PyEnum_Type"
            elif constant is frozenset:
                return "(PyObject *)&PyFrozenSet_Type"
            elif python_version >= 0x270 and constant is memoryview:
                return "(PyObject *)&PyMemoryView_Type"
            elif python_version < 0x300 and constant is basestring:
                return "(PyObject *)&PyBaseString_Type"
            elif python_version < 0x300 and constant is xrange:
                return "(PyObject *)&PyRange_Type"
            elif constant in builtin_anon_values:
                return (
                    "(PyObject *)" + builtin_anon_codes[builtin_anon_values[constant]]
                )
            elif constant in builtin_exception_values_list:
                return "(PyObject *)PyExc_%s" % constant.__name__
            elif constant is ExceptionGroup:
                return "(PyObject *)_PyInterpreterState_GET()->exc_state.PyExc_ExceptionGroup"
            elif constant is BaseExceptionGroup:
                return "(PyObject *)PyExc_BaseExceptionGroup"
            else:
                type_name = constant.__name__

                if constant is int and python_version >= 0x300:
                    type_name = "long"
                elif constant is str:
                    type_name = "string" if python_version < 0x300 else "unicode"

                return "(PyObject *)&Py%s_Type" % type_name.capitalize()
        else:
            return self._getConstantCode(constant)

    def _getConstantCode(self, constant):
        key = "const_" + namifyConstant(constant)

        if key not in self.constants:
            self.constants.add(key)
            self.constants_writer.addConstantValue(constant)

        key = "%s[%d]" % (self.top_level_name, self.constants.index(key))

        # TODO: Make it returning, more clear.
        return key

    def getBlobDataCode(self, data, name):
        key = "blob_" + namifyConstant(data)

        if key not in self.constants:
            self.constants.add(key)
            self.constants_writer.addBlobData(data=data, name=name)

        key = "%s[%d]" % (self.top_level_name, self.constants.index(key))

        return key

    def getConstantsCount(self):
        # Make sure to add no more after asking this.
        self.constants_writer.close()

        return len(self.constants)

    def getConstantNames(self):
        return tuple(self.constants)


class ConstantAccessor(GlobalConstantAccessor):
    __slots__ = ()

    def _getConstantCode(self, constant):
        key = "const_" + namifyConstant(constant)
        if key in self.global_constant_keys:
            return key

        if key not in self.constants:
            self.constants.add(key)
            self.constants_writer.addConstantValue(constant)

        return "%s.%s" % (self.top_level_name, key)

    def getBlobDataCode(self, data, name):
        key = "blob_" + namifyConstant(data)

        if key not in self.constants:
            self.constants.add(key)
            self.constants_writer.addBlobData(data=data, name=name)

        return "%s[%d]" % (self.top_level_name, self.constants.index(key))


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Import related codes.

That is import as expression, and star import.
"""

import os

from nuitka.HardImportRegistry import isHardModule, isHardModuleDynamic
from nuitka.nodes.LocalsScopes import GlobalsDictHandle
from nuitka.PythonVersions import python_version
from nuitka.utils.Jinja2 import renderTemplateFromString
from nuitka.utils.ModuleNames import ModuleName

from .CodeHelpers import (
    generateChildExpressionsCode,
    generateExpressionCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitBoolCode, getErrorExitCode
from .LineNumberCodes import emitLineNumberUpdateCode
from .ModuleCodes import getModuleAccessCode


def generateBuiltinImportCode(to_name, expression, emit, context):
    # We know that 5 expressions are created, pylint: disable=W0632
    (
        module_name,
        globals_name,
        locals_name,
        import_list_name,
        level_name,
    ) = generateChildExpressionsCode(expression=expression, emit=emit, context=context)

    with withObjectCodeTemporaryAssignment(
        to_name, "imported_value", expression, emit, context
    ) as value_name:
        _getBuiltinImportCode(
            expression=expression,
            to_name=value_name,
            module_name=module_name,
            globals_name=globals_name,
            locals_name=locals_name,
            import_list_name=import_list_name,
            level_name=level_name,
            needs_check=expression.mayRaiseException(BaseException),
            emit=emit,
            context=context,
        )


# TODO: Maybe use this for other cases too, not just import.
def _getCountedArgumentsHelperCallCode(
    helper_prefix, to_name, args, min_args, needs_check, emit, context
):
    orig_args = args
    args = list(args)
    while args[-1] is None:
        del args[-1]

    if None in args:
        emit(
            "%s = %s_KW(tstate, %s);"
            % (
                to_name,
                helper_prefix,
                ", ".join("NULL" if arg is None else str(arg) for arg in orig_args),
            )
        )
    else:
        # Check that no following arguments are not None.
        assert len(args) >= min_args

        emit(
            "%s = %s%d(tstate, %s);"
            % (to_name, helper_prefix, len(args), ", ".join(str(arg) for arg in args))
        )

    getErrorExitCode(
        check_name=to_name,
        release_names=args,
        needs_check=needs_check,
        emit=emit,
        context=context,
    )

    context.addCleanupTempName(to_name)


def _getBuiltinImportCode(
    expression,
    to_name,
    module_name,
    globals_name,
    locals_name,
    import_list_name,
    level_name,
    needs_check,
    emit,
    context,
):
    emitLineNumberUpdateCode(expression, emit, context)

    # Spell the resulting helper names out, so this can be found,
    # IMPORT_MODULE_KW, IMPORT_MODULE1, IMPORT_MODULE2, IMPORT_MODULE3, IMPORT_MODULE5
    _getCountedArgumentsHelperCallCode(
        helper_prefix="IMPORT_MODULE",
        to_name=to_name,
        args=(module_name, globals_name, locals_name, import_list_name, level_name),
        min_args=1,
        needs_check=needs_check,
        emit=emit,
        context=context,
    )


def generateImportModuleFixedCode(to_name, expression, emit, context):
    needs_check = expression.mayRaiseException(BaseException)

    if needs_check:
        emitLineNumberUpdateCode(expression, emit, context)

    with withObjectCodeTemporaryAssignment(
        to_name, "imported_value", expression, emit, context
    ) as value_name:
        emit(
            """%s = IMPORT_MODULE_FIXED(tstate, %s, %s);"""
            % (
                value_name,
                context.getConstantCode(expression.getModuleName().asString()),
                context.getConstantCode(expression.getValueName().asString()),
            )
        )

        getErrorExitCode(
            check_name=value_name, needs_check=needs_check, emit=emit, context=context
        )

        context.addCleanupTempName(value_name)


def getImportModuleHardCodeName(module_name):
    """Encoding hard module name for code name."""

    module_name = ModuleName(module_name)

    return "IMPORT_HARD_%s" % module_name.asPath().replace(os.path.sep, "__").upper()


def generateImportModuleHardCode(to_name, expression, emit, context):
    imported_module_name = expression.getModuleName()
    module_value_name = expression.getValueName()

    needs_check = expression.mayRaiseException(BaseException)

    if needs_check:
        emitLineNumberUpdateCode(expression, emit, context)

    with withObjectCodeTemporaryAssignment(
        to_name, "imported_value", expression, emit, context
    ) as value_name:
        import_gives_ref, module_getter_code = getImportHardModuleGetterCode(
            module_name=imported_module_name, context=context
        )

        if imported_module_name == module_value_name:
            emit("""%s = %s;""" % (value_name, module_getter_code))
        else:
            import_gives_ref, module_getter_code2 = getImportHardModuleGetterCode(
                module_name=module_value_name, context=context
            )

            emit(
                renderTemplateFromString(
                    r"""
{% if import_gives_ref %}
{
    PyObject *hard_module = {{module_getter_code1}};
    Py_DECREF(hard_module);
}
{% else %}
{{module_getter_code1}};
{% endif %}
{{value_name}} = {{module_getter_code2}};
""",
                    value_name=value_name,
                    module_getter_code1=module_getter_code,
                    module_getter_code2=module_getter_code2,
                    import_gives_ref=import_gives_ref,
                )
            )

        getErrorExitCode(
            check_name=value_name, needs_check=needs_check, emit=emit, context=context
        )

        if import_gives_ref:
            context.addCleanupTempName(value_name)


def generateConstantSysVersionInfoCode(to_name, expression, emit, context):
    with withObjectCodeTemporaryAssignment(
        to_name, "imported_value", expression, emit, context
    ) as value_name:
        emit("""%s = Nuitka_SysGetObject("%s");""" % (value_name, "version_info"))

    getErrorExitCode(
        check_name=value_name, needs_check=False, emit=emit, context=context
    )


def getImportHardModuleGetterCode(module_name, context):
    if isHardModuleDynamic(module_name):
        module_name_code = context.getConstantCode(module_name.asString())

        module_getter_code = "IMPORT_MODULE_FIXED(tstate, %s, %s)" % (
            module_name_code,
            module_name_code,
        )
        gives_ref = True
    else:
        module_getter_code = "%s()" % getImportModuleHardCodeName(module_name)
        gives_ref = False

    return gives_ref, module_getter_code


def getImportModuleNameHardCode(
    to_name, module_name, import_name, needs_check, emit, context
):
    module_name = ModuleName(module_name)

    if module_name == "sys":
        emit("""%s = Nuitka_SysGetObject("%s");""" % (to_name, import_name))
        needs_release = False
    elif isHardModule(module_name):
        if needs_check:
            emitLineNumberUpdateCode(expression=None, emit=emit, context=context)

        import_gives_ref, module_getter_code = getImportHardModuleGetterCode(
            module_name=module_name, context=context
        )

        emit(
            renderTemplateFromString(
                r"""
{
    PyObject *hard_module = {{module_getter_code}};
{% if needs_check %}
    if (likely(hard_module != NULL)) {
        {{to_name}} = LOOKUP_ATTRIBUTE(tstate, hard_module, {{import_name}});

{% if import_gives_ref %}
        Py_DECREF(hard_module);
{% endif %}

    } else {
        {{to_name}} = NULL;
    }
{% else %}
    {{to_name}} = LOOKUP_ATTRIBUTE(tstate, hard_module, {{import_name}});
{% if import_gives_ref %}
    Py_DECREF(hard_module);
{% endif %}
{% endif %}
}
""",
                to_name=to_name,
                module_name=str(module_name),
                module_getter_code=module_getter_code,
                import_name=context.getConstantCode(import_name),
                needs_check=needs_check,
                import_gives_ref=import_gives_ref,
            )
        )

        needs_release = True
    else:
        assert False, module_name

    getErrorExitCode(
        check_name=to_name, needs_check=needs_check, emit=emit, context=context
    )

    if needs_release:
        context.addCleanupTempName(to_name)


def generateImportModuleNameHardCode(to_name, expression, emit, context):
    with withObjectCodeTemporaryAssignment(
        to_name, "imported_value", expression, emit, context
    ) as value_name:
        context.setCurrentSourceCodeReference(expression.getCompatibleSourceReference())

        getImportModuleNameHardCode(
            to_name=value_name,
            module_name=expression.getModuleName(),
            import_name=expression.getImportName(),
            needs_check=expression.mayRaiseException(BaseException),
            emit=emit,
            context=context,
        )


def generateImportlibImportCallCode(to_name, expression, emit, context):
    needs_check = expression.mayRaiseException(BaseException)

    with withObjectCodeTemporaryAssignment(
        to_name, "imported_module", expression, emit, context
    ) as value_name:
        import_name, package_name = generateChildExpressionsCode(
            expression=expression, emit=emit, context=context
        )

        emitLineNumberUpdateCode(expression, emit, context)

        # TODO: The import name wouldn't have to be an object really, could do with a
        # C string only.

        emit(
            renderTemplateFromString(
                r"""
{
    PyObject *hard_module = {{import_hard_importlib}}();
    PyObject *import_module_func = LOOKUP_ATTRIBUTE(tstate, hard_module, {{context.getConstantCode("import_module")}});
{% if package_name == None %}
    {{to_name}} = CALL_FUNCTION_WITH_SINGLE_ARG(tstate, import_module_func, {{import_name}});
{% else %}
    PyObject *args[2] = { {{import_name}}, {{package_name}} };
    {{to_name}} = CALL_FUNCTION_WITH_ARGS2(tstate, import_module_func, args);
{% endif %}
    Py_DECREF(import_module_func);
}
""",
                context=context,
                to_name=value_name,
                import_name=import_name,
                package_name=package_name,
                import_hard_importlib=getImportModuleHardCodeName("importlib"),
            )
        )

        getErrorExitCode(
            check_name=value_name,
            release_names=(import_name, package_name),
            needs_check=needs_check,
            emit=emit,
            context=context,
        )


def generateImportStarCode(statement, emit, context):
    module_name = context.allocateTempName("star_imported")

    generateExpressionCode(
        to_name=module_name,
        expression=statement.subnode_module,
        emit=emit,
        context=context,
    )

    with context.withCurrentSourceCodeReference(statement.getSourceReference()):
        res_name = context.getBoolResName()

        target_scope = statement.getTargetDictScope()

        if type(target_scope) is GlobalsDictHandle:
            emit(
                "%s = IMPORT_MODULE_STAR(tstate, %s, true, %s);"
                % (res_name, getModuleAccessCode(context=context), module_name)
            )
        else:
            locals_declaration = context.addLocalsDictName(target_scope.getCodeName())

            emit(
                "%(res_name)s = IMPORT_MODULE_STAR(tstate, %(locals_dict)s, false, %(module_name)s);"
                % {
                    "res_name": res_name,
                    "locals_dict": locals_declaration,
                    "module_name": module_name,
                }
            )

        getErrorExitBoolCode(
            condition="%s == false" % res_name,
            release_name=module_name,
            emit=emit,
            context=context,
        )


def generateImportNameCode(to_name, expression, emit, context):
    from_arg_name = context.allocateTempName("import_name_from")

    generateExpressionCode(
        to_name=from_arg_name,
        expression=expression.subnode_module,
        emit=emit,
        context=context,
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "imported_value", expression, emit, context
    ) as value_name:
        if python_version >= 0x350:
            emit(
                """\
if (PyModule_Check(%(from_arg_name)s)) {
    %(to_name)s = IMPORT_NAME_OR_MODULE(
        tstate,
        %(from_arg_name)s,
        (PyObject *)moduledict_%(module_identifier)s,
        %(import_name)s,
        %(import_level)s
    );
} else {
    %(to_name)s = IMPORT_NAME_FROM_MODULE(tstate, %(from_arg_name)s, %(import_name)s);
}
"""
                % {
                    "to_name": value_name,
                    "from_arg_name": from_arg_name,
                    "import_name": context.getConstantCode(
                        constant=expression.getImportName()
                    ),
                    "import_level": context.getConstantCode(
                        constant=expression.getImportLevel()
                    ),
                    "module_identifier": context.getModuleCodeName(),
                }
            )
        else:
            emit(
                "%s = IMPORT_NAME_FROM_MODULE(tstate, %s, %s);"
                % (
                    value_name,
                    from_arg_name,
                    context.getConstantCode(constant=expression.getImportName()),
                )
            )

        getErrorExitCode(
            check_name=value_name,
            release_name=from_arg_name,
            needs_check=expression.mayRaiseException(BaseException),
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(value_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Finalizations. Last steps directly before code creation is called.

Here the final tasks are executed. Things normally volatile during optimization
can be computed here, so the code generation can be quick and doesn't have to
check it many times.

"""

from nuitka.tree import Operations

from .FinalizeMarkups import FinalizeMarkups


def prepareCodeGeneration(module):
    visitor = FinalizeMarkups(module)
    Operations.visitTree(module, visitor)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



"""This module is only an abstraction of OrderedSet which is not present in
Python at all.

spell-checker: ignore orderedset,orderedsets

"""

from nuitka.PythonVersions import python_version

recommended_orderedset_package_name, recommended_orderedset_module_name = (
    ("ordered-set", "ordered_set")
    if python_version >= 0x370
    else ("orderedset", "orderedsets")
)


def _tryImportOrderedSet():
    try:
        from orderedset import OrderedSet as result
    except ImportError:
        try:
            from ordered_set import OrderedSet as result
        except ImportError:
            return None

    return result


OrderedSet = _tryImportOrderedSet()


if OrderedSet is None:
    from .OrderedSetsFallback import OrderedSet


def buildOrderedSet(*producers):
    """Helper function to merge multiple producers into one OrderedSet value"""
    values = []

    for producer in producers:
        values.extend(producer)

    return OrderedSet(values)



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Tuple codes"""

from .CodeHelpers import (
    decideConversionCheckNeeded,
    generateExpressionCode,
    withCleanupFinally,
    withObjectCodeTemporaryAssignment,
)
from .PythonAPICodes import generateCAPIObjectCode


def _areConstants(expressions):
    for expression in expressions:
        if not expression.isExpressionConstantRef():
            return False

        if expression.isMutable():
            return False

    return True


def generateTupleCreationCode(to_name, expression, emit, context):
    with withObjectCodeTemporaryAssignment(
        to_name, "tuple_value", expression, emit, context
    ) as value_name:
        getTupleCreationCode(
            to_name=value_name,
            elements=expression.subnode_elements,
            emit=emit,
            context=context,
        )


def getTupleCreationCode(to_name, elements, emit, context):
    if _areConstants(elements):
        to_name.getCType().emitAssignmentCodeFromConstant(
            to_name=to_name,
            constant=tuple(element.getCompileTimeConstant() for element in elements),
            # TODO: Would depend on our target being escaping.
            may_escape=True,
            emit=emit,
            context=context,
        )
    else:
        element_name = context.allocateTempName("tuple_element")

        def generateElementCode(element):
            generateExpressionCode(
                to_name=element_name, expression=element, emit=emit, context=context
            )

            # Use helper that makes sure we provide a reference.
            if context.needsCleanup(element_name):
                context.removeCleanupTempName(element_name)
                helper_code = "PyTuple_SET_ITEM"
            else:
                helper_code = "PyTuple_SET_ITEM0"

            return helper_code

        helper_code = generateElementCode(elements[0])

        # TODO: Consider more compact tuple creation helpers
        emit("%s = MAKE_TUPLE_EMPTY(tstate, %d);" % (to_name, len(elements)))

        needs_exception_exit = any(
            element.mayRaiseException(BaseException) for element in elements[1:]
        )

        with withCleanupFinally(
            "tuple_build", to_name, needs_exception_exit, emit, context
        ) as guarded_emit:
            emit = guarded_emit.emit

            for count, element in enumerate(elements):
                if count > 0:
                    helper_code = generateElementCode(element)

                emit("%s(%s, %d, %s);" % (helper_code, to_name, count, element_name))


def generateBuiltinTupleCode(to_name, expression, emit, context):
    generateCAPIObjectCode(
        to_name=to_name,
        capi="PySequence_Tuple",
        tstate=False,
        arg_desc=(("tuple_arg", expression.subnode_value),),
        may_raise=expression.mayRaiseException(BaseException),
        conversion_check=decideConversionCheckNeeded(to_name, expression),
        source_ref=expression.getCompatibleSourceReference(),
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""For macOS DMG file creation"""

import os

from nuitka.OutputDirectories import (
    getResultBasePath,
    getStandaloneDirectoryPath,
)
from nuitka.utils.Execution import (
    executeToolChecked,
    filterOutputByLine,
    getExecutablePath,
)
from nuitka.utils.FileOperations import changeFilenameExtension, deleteFile
from nuitka.utils.Signing import addMacOSCodeSignature

from .MacOSApp import getMacOSIconPaths


def _filterCreateDmgOutput(stderr):
    def isNonErrorExit(line):
        if b".DS_STORE to be created" in line:
            return True
        if b"Searching for mounted" in line:
            return True

        return False

    return filterOutputByLine(stderr, isNonErrorExit)


def createDmgFile(logger):
    """Create a DMG file for the application bundle."""
    create_dmg_path = getExecutablePath("termux-elf-cleaner")

    if create_dmg_path is None and os.path.exists("/opt/homebrew/bin/create-dmg"):
        create_dmg_path = "/opt/homebrew/bin/create-dmg"

    # TODO: Move that to options checking.
    if create_dmg_path is None:
        return logger.sysexit("Cannot find 'create-dmg' tool, not creating DMG.")

    app_bundle_path = getStandaloneDirectoryPath(bundle=False, real=True)

    dmg_path = changeFilenameExtension(app_bundle_path, ".dmg")
    deleteFile(dmg_path, must_exist=False)

    # spell-checker: ignore volname,volicon
    command = [
        create_dmg_path,
        "--volname",
        os.path.basename(getResultBasePath()),
        "--window-pos",
        "200",
        "120",
        "--window-size",
        "800",
        "400",
        "--icon-size",
        "100",
        "--icon",
        os.path.basename(app_bundle_path),
        "200",
        "190",
        "--hide-extension",
        os.path.basename(app_bundle_path),
        "--app-drop-link",
        "600",
        "185",
        dmg_path,
        app_bundle_path,
    ]

    icon_paths = getMacOSIconPaths()
    if icon_paths:
        command.extend(["--volicon", icon_paths[0]])

    logger.info("Creating DMG file '%s'..." % dmg_path)

    executeToolChecked(
        logger=logger,
        command=command,
        absence_message="Cannot find 'create-dmg' tool necessary to create DMG.",
        stderr_filter=_filterCreateDmgOutput,
    )

    addMacOSCodeSignature(filenames=[dmg_path], entitlements_filename=None)

    if os.path.exists(dmg_path):
        logger.info(
            "Created DMG file with the application bundle in it: '%s'." % dmg_path
        )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Standard plug-in to tell user about needed or useful options for packages.

When certain GUI packages are used, disabling the console may or may not be what
the user wants, or even be required, as e.g. "wx" on macOS will crash unless the
console is disabled. This reads Yaml configuration.
"""

from nuitka.options.Options import (
    isOnefileMode,
    isStandaloneMode,
    shallCreateAppBundle,
)
from nuitka.plugins.YamlPluginBase import NuitkaYamlPluginBase
from nuitka.utils.Utils import isMacOS


class NuitkaPluginOptionsNanny(NuitkaYamlPluginBase):
    plugin_name = "options-nanny"
    plugin_desc = (
        "Inform user about potential problems as per package configuration files."
    )
    plugin_category = "core,package-support"

    @staticmethod
    def isAlwaysEnabled():
        return True

    def sysexitIllegalOptionValue(self, full_name, option, value):
        self.sysexit(
            "Illegal value for package '%s' option '%s' ('%s')"
            % (full_name, option, value)
        )

    def _checkSupportedVersion(self, full_name, support_info, description, condition):
        # Configured to not report, default value.
        if support_info == "ignore":
            return

        # No issues at all, but remember that for code generation.
        if support_info == "perfect":
            from nuitka.code_generation.LoaderCodes import (
                markModuleAsPerfectSupported,
            )

            markModuleAsPerfectSupported(full_name)
            return

        if support_info == "parameter":
            message = "Module '%s' has parameter: %s" % (full_name, description)
        elif support_info == "plugin":
            message = "Module '%s' has plugin consideration: %s" % (
                full_name,
                description,
            )
        else:
            if condition != "True":
                problem_desc = (
                    " with incomplete support due to condition '%s'" % condition
                )
            else:
                problem_desc = " with incomplete support"

            message = "Using module '%s' (version %s)%s: %s" % (
                full_name,
                ".".join(str(d) for d in self.getPackageVersion(full_name)),
                problem_desc,
                description,
            )

        if support_info == "error":
            self.sysexit(message)
        elif support_info in ("warning", "parameter", "plugin"):
            self.warning(message)
        elif support_info == "info":
            self.info(message)
        else:
            self.sysexit(
                "Error, unknown support_info level '%s' for module '%s'"
                % (support_info, full_name.asString())
            )

    def _checkMacOSBundleMode(self, full_name, macos_bundle):
        if macos_bundle == "yes":
            if isStandaloneMode() and not shallCreateAppBundle():
                self.sysexit(
                    """\
Error, package '%s' requires '--mode=app' to be used or else it cannot work."""
                    % full_name
                )
        elif macos_bundle == "no":
            pass
        elif macos_bundle == "recommend":
            if not shallCreateAppBundle():
                self.info(
                    """\
Note, when using '%s', consider using '--mode=app' option. Otherwise \
high resolution will not be available and a terminal window will open. \
However for debugging, terminal output is the easiest way to see \
informative traceback and error information, so launch it from there if \
possible."""
                    % full_name
                )
        else:
            self.sysexitIllegalOptionValue(full_name, "macos_bundle", macos_bundle)

    def _checkMacOSBundleOnefileMode(self, full_name, macos_bundle_as_onefile):
        if macos_bundle_as_onefile == "yes":
            if isStandaloneMode() and shallCreateAppBundle() and not isOnefileMode():
                self.sysexit(
                    """\
Error, package '%s' requires '--onefile' to be used on top of '--macos-create-app-bundle' or else it cannot work."""
                    % full_name
                )
        elif macos_bundle_as_onefile == "no":
            pass
        else:
            self.sysexitIllegalOptionValue(
                full_name, "macos_bundle_onefile_mode", macos_bundle_as_onefile
            )

    # TODO: Definitely the wrong function to use, but we migrated this out of
    # implicit imports, where it was done there.
    def getImplicitImports(self, module):
        full_name = module.getFullName()

        for options_config in self.config.get(full_name, section="options"):
            for check in options_config.get("checks", ()):
                condition = check.get("when", "True")

                if self.evaluateCondition(full_name=full_name, condition=condition):
                    self._checkSupportedVersion(
                        full_name=full_name,
                        support_info=check.get("support_info", "ignore"),
                        description=check.get("description", "not given"),
                        condition=condition,
                    )

                    if isMacOS():
                        self._checkMacOSBundleMode(
                            full_name=full_name,
                            macos_bundle=check.get("macos_bundle", "no"),
                        )

                        self._checkMacOSBundleOnefileMode(
                            full_name=full_name,
                            macos_bundle_as_onefile=check.get(
                                "macos_bundle_as_onefile", "no"
                            ),
                        )

        return ()


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



"""SCons.Platform.posix

Platform-specific initialization for POSIX (Linux, UNIX, etc.) systems.

There normally shouldn't be any need to import this module directly.  It
will usually be imported through the generic SCons.Platform.Platform()
selection method.
"""

#
# Copyright (c) 2001 - 2014 The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Platform/posix.py  2014/07/05 09:42:21 garyo"

import errno
import os
import os.path
import subprocess
import sys
import select

import SCons.Util
from SCons.Platform import TempFileMunge

exitvalmap = {
    2 : 127,
    13 : 126,
}

def escape(arg):
    "escape shell special characters"
    slash = '\\'
    special = '"$()'

    arg = arg.replace(slash, slash+slash)
    for c in special:
        arg = arg.replace(c, slash+c)

    return '"' + arg + '"'

def exec_subprocess(l, env):
    proc = subprocess.Popen(l, env = env, close_fds = True)
    return proc.wait()

def subprocess_spawn(sh, escape, cmd, args, env):
    return exec_subprocess([sh, '-c', ' '.join(args)], env)

def exec_popen3(l, env, stdout, stderr):
    proc = subprocess.Popen(l, env = env, close_fds = True,
                            stdout = stdout,
                            stderr = stderr)
    return proc.wait()

def piped_env_spawn(sh, escape, cmd, args, env, stdout, stderr):
    # spawn using Popen3 combined with the env command
    # the command name and the command's stdout is written to stdout
    # the command's stderr is written to stderr
    return exec_popen3([sh, '-c', ' '.join(args)],
                       env, stdout, stderr)


def generate(env):
    # Bearing in mind we have python 2.4 as a baseline, we can just do this:
    spawn = subprocess_spawn
    pspawn = piped_env_spawn
    # Note that this means that 'escape' is no longer used

    if 'ENV' not in env:
        env['ENV']        = {}
    env['ENV']['PATH']    = '/usr/local/bin:/opt/bin:/bin:/usr/bin'
    env['OBJPREFIX']      = ''
    env['OBJSUFFIX']      = '.o'
    env['SHOBJPREFIX']    = '$OBJPREFIX'
    env['SHOBJSUFFIX']    = '$OBJSUFFIX'
    env['PROGPREFIX']     = ''
    env['PROGSUFFIX']     = ''
    env['LIBPREFIX']      = 'lib'
    env['LIBSUFFIX']      = '.a'
    env['SHLIBPREFIX']    = '$LIBPREFIX'
    env['SHLIBSUFFIX']    = '.so'
    env['LIBPREFIXES']    = [ '$LIBPREFIX' ]
    env['LIBSUFFIXES']    = [ '$LIBSUFFIX', '$SHLIBSUFFIX' ]
    env['PSPAWN']         = pspawn
    env['SPAWN']          = spawn
    env['SHELL']          = 'sh'
    env['ESCAPE']         = escape
    env['TEMPFILE']       = TempFileMunge
    env['TEMPFILEPREFIX'] = '@'
    #Based on LINUX: ARG_MAX=ARG_MAX=131072 - 3000 for environment expansion
    #Note: specific platforms might rise or lower this value
    env['MAXLINELENGTH']  = 128072

    # This platform supports RPATH specifications.
    env['__RPATH'] = '$_RPATH'

    # GDC is GCC family, but DMD and LDC have different options.
    # Must be able to have GCC and DMD work in the same build, so:
    env['__DRPATH'] = '$_DRPATH'

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Variables link the storage and use of a Python variable together.

Different kinds of variables represent different scopes and owners types,
and their links between each other, i.e. references as in closure or
module variable references.

"""

from abc import abstractmethod

from nuitka.__past__ import iterItems
from nuitka.nodes.shapes.BuiltinTypeShapes import tshape_dict
from nuitka.nodes.shapes.StandardShapes import tshape_unknown
from nuitka.utils.CStrings import encodePythonIdentifierToC
from nuitka.utils.InstanceCounters import (
    counted_del,
    counted_init,
    isCountingInstances,
)
from nuitka.utils.SlotMetaClasses import getMetaClassBase

complete = False


class Variable(getMetaClassBase("Variable", require_slots=True)):
    # We will need all of these attributes, since we track the global
    # state and cache some decisions as attributes. TODO: But in some
    # cases, part of the these might be moved to the outside.
    __slots__ = (
        "variable_name",
        "owner",
        "version_number",
        "shared_users",
        "traces",
        "users",
        "writers",
    )

    @counted_init
    def __init__(self, owner, variable_name):
        # assert type(variable_name) is str, variable_name
        # assert type(owner) not in (tuple, list), owner

        self.variable_name = variable_name
        self.owner = owner

        self.version_number = 0

        self.shared_users = False

        self.traces = {}

        # Derived from all traces.
        self.writers = set()

    if isCountingInstances():
        __del__ = counted_del()

    def finalize(self):
        del self.writers
        del self.traces
        del self.owner

    def __repr__(self):
        return "<%s '%s' of '%s'>" % (
            self.__class__.__name__,
            self.variable_name,
            self.owner.getName(),
        )

    @abstractmethod
    def getVariableType(self):
        pass

    def getDescription(self):
        return "variable '%s'" % self.variable_name

    def getName(self):
        return self.variable_name

    def getOwner(self):
        return self.owner

    def getEntryPoint(self):
        return self.owner.getEntryPoint()

    def getVariableCodeName(self):
        return encodePythonIdentifierToC(self.variable_name)

    def allocateTargetNumber(self):
        self.version_number += 3

        return self.version_number

    @staticmethod
    def isLocalVariable():
        return False

    @staticmethod
    def isParameterVariable():
        return False

    @staticmethod
    def isModuleVariable():
        return False

    @staticmethod
    def isIncompleteModuleVariable():
        return False

    @staticmethod
    def isTempVariable():
        return False

    @staticmethod
    def isTempVariableBool():
        return False

    @staticmethod
    def isLocalsDictVariable():
        return False

    def addVariableUser(self, user):
        # Update the shared scopes flag.
        if user is not self.owner:
            self.shared_users = True

            # These are not really scopes, just shared uses.
            if (
                user.isExpressionGeneratorObjectBody()
                or user.isExpressionCoroutineObjectBody()
                or user.isExpressionAsyncgenObjectBody()
            ):
                if self.owner is user.getParentVariableProvider():
                    return

            _variables_in_shared_scopes.add(self)

    def isSharedTechnically(self):
        if not self.shared_users:
            return False

        if not self.traces:
            return False

        owner = self.owner.getEntryPoint()

        for user in self.traces:
            user = user.getEntryPoint()

            while user is not owner and (
                (user.isExpressionFunctionBody() and not user.needsCreation())
                or user.isExpressionClassBodyBase()
            ):
                user = user.getParentVariableProvider()

            if user is not owner:
                return True

        return False

    def setTracesForUserFirst(self, user, variable_traces):
        self.traces[user] = variable_traces

        for trace in variable_traces.values():
            if trace.isAssignTrace():
                self.writers.add(user)

                break
            if user is not self.owner and trace.isDeletedTrace():
                self.writers.add(user)

                break

    def setTracesForUserUpdate(self, user, variable_traces):
        self.traces[user] = variable_traces

        if user in self.writers:
            for trace in variable_traces.values():
                if trace.isAssignTrace():
                    break
                if user is not self.owner and trace.isDeletedTrace():
                    break
            else:
                self.writers.remove(user)

    def removeTracesForUser(self, user):
        del self.traces[user]

        if user in self.writers:
            self.writers.remove(user)

    def hasEmptyTracesFor(self, user):
        """Do these traces contain any usage."""
        if user in self.traces:
            for trace in self.traces[user].values():
                if trace.isUsingTrace():
                    return False

        return True

    def hasNoWritingTraces(self):
        """Do these traces contain any writes."""

        for traces in self.traces.values():
            for trace in traces.values():
                if trace.isWritingTrace():
                    return False

        return True

    def hasAccessesOutsideOf(self, provider):
        if not self.owner.locals_scope.complete:
            return None
        elif not self.traces:
            return False
        elif provider in self.traces:
            return len(self.traces) > 1
        else:
            return True

    def hasWritersOutsideOf(self, provider):
        if not self.owner.locals_scope.complete:
            # TODO: Maybe this doesn't have to be limited to these types.
            if not self.shared_users and (
                self.isLocalVariable() or self.isTempVariable()
            ):
                return False
            return None
        elif not self.writers:
            return False
        elif provider in self.writers:
            return len(self.writers) > 1
        else:
            return True

    def getMatchingUnescapedAssignTrace(self, assign_node):
        found = None
        for traces in self.traces.values():
            for trace in traces.values():
                if trace.isAssignTrace():
                    if trace.getAssignNode() is assign_node:
                        found = trace
                elif trace.isEscapeTrace():
                    return None

            if found is not None:
                return found

        return found

    def getTypeShapes(self):
        result = set()

        for traces in self.traces.values():
            for trace in traces.values():
                if trace.isAssignTrace():
                    result.add(trace.getAssignNode().getTypeShape())
                elif trace.isUnknownTrace():
                    result.add(tshape_unknown)
                elif trace.isEscapeTrace():
                    result.add(tshape_unknown)
                elif trace.isInitTrace():
                    result.add(tshape_unknown)
                elif trace.isUnassignedTrace():
                    pass
                elif trace.isMergeTrace():
                    pass
                # TODO: Remove this and be not unknown.
                elif trace.isLoopTrace():
                    trace.getTypeShape().emitAlternatives(result.add)
                else:
                    assert False, trace

        return result

    @staticmethod
    def onControlFlowEscape(trace_collection):
        """Mark the variable as escaped or unknown, or keep it depending on variable type."""

    def removeKnowledge(self, trace_collection):
        """Remove knowledge for the variable marking as unknown or escaped."""
        trace_collection.markActiveVariableAsEscaped(self)

    def removeAllKnowledge(self, trace_collection):
        """Remove all knowledge for the variable marking as unknown, or keep it depending on variable type."""
        trace_collection.markActiveVariableAsUnknown(self)


class LocalVariable(Variable):
    __slots__ = ()

    def __init__(self, owner, variable_name):
        Variable.__init__(self, owner=owner, variable_name=variable_name)

    def makeClone(self, new_owner):
        return LocalVariable(owner=new_owner, variable_name=self.variable_name)

    @staticmethod
    def isLocalVariable():
        return True

    def initVariableLate(self, trace_collection):
        """Initialize variable in trace collection state."""
        trace_collection.variable_escapable.add(self)
        trace_collection.has_unescaped_variables = True
        return trace_collection.initVariableUninitialized(self, None)

    if str is not bytes:

        def onControlFlowEscape(self, trace_collection):
            if self.hasWritersOutsideOf(trace_collection.owner) is not False:
                trace_collection.markClosureVariableAsUnknown(self)
            elif self.hasAccessesOutsideOf(trace_collection.owner) is not False:
                trace_collection.markActiveVariableAsEscaped(self)

    else:

        def onControlFlowEscape(self, trace_collection):
            if self.hasAccessesOutsideOf(trace_collection.owner) is not False:
                trace_collection.markActiveVariableAsEscaped(self)

    @staticmethod
    def getVariableType():
        return "object"


class ParameterVariable(LocalVariable):
    __slots__ = ()

    def __init__(self, owner, parameter_name):
        LocalVariable.__init__(self, owner=owner, variable_name=parameter_name)

    def makeClone(self, new_owner):
        return ParameterVariable(owner=new_owner, parameter_name=self.variable_name)

    def getDescription(self):
        return "parameter variable '%s'" % self.variable_name

    @staticmethod
    def isParameterVariable():
        return True

    def initVariableLate(self, trace_collection):
        """Initialize variable in trace collection state."""
        return trace_collection.initVariableInit(self, None)


class ModuleVariable(Variable):
    __slots__ = ()

    def __init__(self, module, variable_name):
        # assert type(variable_name) is str, repr(variable_name)
        # assert module.isCompiledPythonModule()

        Variable.__init__(self, owner=module, variable_name=variable_name)

    def __repr__(self):
        return "<ModuleVariable '%s' of '%s'>" % (
            self.variable_name,
            self.owner.getFullName(),
        )

    def getDescription(self):
        return "global variable '%s'" % self.variable_name

    @staticmethod
    def isModuleVariable():
        return True

    def initVariableLate(self, trace_collection):
        """Initialize variable in trace collection state."""
        trace_collection.variable_escapable.add(self)
        trace_collection.has_unescaped_variables = True
        return trace_collection.initVariableModule(self, None)

    def onControlFlowEscape(self, trace_collection):
        trace_collection.markActiveVariableAsUnknown(self)

    def removeKnowledge(self, trace_collection):
        """Remove knowledge for the variable marking as unknown or escaped."""
        trace_collection.markActiveVariableAsUnknown(self)

    def isIncompleteModuleVariable(self):
        return not self.owner.locals_scope.complete

    def hasDefiniteWrites(self):
        if not self.owner.locals_scope.complete:
            return None
        else:
            return bool(self.writers)

    def getModule(self):
        return self.owner

    @staticmethod
    def getVariableType():
        return "object"


class TempVariable(Variable):
    __slots__ = ("variable_type",)

    def __init__(self, owner, variable_name, variable_type):
        Variable.__init__(self, owner=owner, variable_name=variable_name)

        # TODO: Push this later down to Variable itself.
        self.variable_type = variable_type

    @staticmethod
    def isTempVariable():
        return True

    def getVariableType(self):
        return self.variable_type

    def isTempVariableBool(self):
        return self.variable_type == "bool"

    def getDescription(self):
        return "temp variable '%s'" % self.variable_name

    def initVariableLate(self, trace_collection):
        """Initialize variable in trace collection state."""
        return trace_collection.initVariableUninitialized(self, None)

    @staticmethod
    def removeAllKnowledge(trace_collection):
        """Remove all knowledge for the variable marking as unknown, or keep it depending on variable type."""
        # For temporary variables, the knowledge is not by name, so never gets
        # lost to outside star imports or exec/eval uses.


class LocalsDictVariable(Variable):
    __slots__ = ()

    def __init__(self, owner, variable_name):
        Variable.__init__(self, owner=owner, variable_name=variable_name)

    @staticmethod
    def isLocalsDictVariable():
        return True

    @staticmethod
    def getVariableType():
        return "object"

    def initVariableLate(self, trace_collection):
        """Initialize variable in trace collection state."""
        if self.owner.getTypeShape() is tshape_dict:
            return trace_collection.initVariableUninitialized(self, None)
        else:
            return trace_collection.initVariableUnknown(self, None)

    def inhibitsClassScopeForwardPropagation(self):
        for traces in self.traces.values():
            for trace in traces.values():
                if trace.inhibitsClassScopeForwardPropagation():
                    return True

        return False


def _updateVariablesFromCollectionFirst(new_collection):
    for variable, variable_traces in iterItems(new_collection.getVariableTracesAll()):
        variable.setTracesForUserFirst(new_collection.owner, variable_traces)

    # Release the memory, and prevent the "active" state from being ever
    # inspected, it's useless now.
    new_collection.variable_actives.clear()
    del new_collection.variable_actives


def updateVariablesFromCollection(old_collection, new_collection, source_ref):

    if old_collection is None:
        return _updateVariablesFromCollectionFirst(new_collection)

    old_traces = old_collection.getVariableTracesAll()
    new_traces = new_collection.getVariableTracesAll()
    owner = new_collection.owner
    # Release the memory, and prevent the "active" state from being ever
    # inspected, it's useless now.
    new_collection.variable_actives.clear()
    del new_collection.variable_actives

    for variable, variable_traces in iterItems(new_traces):
        variable.setTracesForUserUpdate(owner, variable_traces)

    for variable in old_traces:
        # Remove traces for variables that are not in the new collection unless
        # they are finalized, then we don't need to update them.
        if variable not in new_traces and hasattr(variable, "users"):
            variable.removeTracesForUser(owner)

    if old_collection.loop_variables != new_collection.loop_variables:
        new_collection.signalChange(
            "var_usage",
            source_ref,
            lambda: "Loop variable '%s' usage ceased."
            % ",".join(
                sorted(
                    variable.getName()
                    for variable in (
                        old_collection.loop_variables - new_collection.loop_variables
                    )
                )
            ),
        )


def removeVariablesFromCollection(old_collection):
    if old_collection is not None:
        owner = old_collection.owner

        for variable in old_collection.getVariableTracesAll():
            variable.removeTracesForUser(owner)


# To detect the Python2 shared variable deletion, that would be a syntax
# error
_variables_in_shared_scopes = set()


def isSharedAmongScopes(variable):
    return variable in _variables_in_shared_scopes


def releaseSharedScopeInformation(tree):
    # Singleton, pylint: disable=global-statement

    assert tree.isCompiledPythonModule()

    global _variables_in_shared_scopes
    _variables_in_shared_scopes = set(
        variable
        for variable in _variables_in_shared_scopes
        if variable.getOwner().getParentModule() is not tree
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Tools for profiling.

Currently using perf kernel interface for native Linux are the only thing exposed,
they don't work inside of containers though, or with default security settings.

Execute that setting permanent weaker:
echo 'kernel.perf_event_paranoid = 1' | sudo tee /etc/sysctl.d/99-nuitka-perf.conf
sudo sysctl --system

Temporary:
sudo sh -c 'echo 1 > /proc/sys/kernel/perf_event_paranoid'

"""

from .Utils import isLinux

# isort: start

if not isLinux():
    # TODO: Windows and MacOS support need other solutions.

    def hasPerfProfilingSupport():
        return False

    PerfCounters = None

else:
    import ctypes
    import ctypes.util
    import fcntl  # Linux only, pylint: disable=I0021,import-error
    import os
    import platform
    import struct

    def hasPerfProfilingSupport():
        return isLinux() and getPerfFileHandle(PERF_COUNT_HW_INSTRUCTIONS) is not None

    # The syscall number for perf_event_open varies by architecture.
    ARCH_MAP = {
        "x86_64": 298,
        "i386": 336,
        "aarch64": 241,
        "armv7l": 364,
    }

    def getPerfFileHandle(config, group_fd=-1, read_format=0):
        SYS_perf_event_open = ARCH_MAP.get(platform.machine())

        if SYS_perf_event_open is None:
            # This is not a hard error, we just cannot do perf.
            return None

        # --- Load libc and define syscall function ---
        libc = ctypes.CDLL(ctypes.util.find_library("c"), use_errno=True)

        syscall = libc.syscall
        syscall.restype = ctypes.c_long

        attr = PerfEventAttr(
            type=PERF_TYPE_HARDWARE,
            size=ctypes.sizeof(PerfEventAttr),
            config=config,
            read_format=read_format,
            # flags for perf_event_open syscall:
            # Bit 0: disabled -> if set start in disabled state
            # Bit 17: exclude_kernel -> if set only count user space
            # Bit 18: exclude_hv -> if set exclude hypervisor
            flags=(1 << 0) | (1 << 17) | (1 << 18),
        )

        # We must explicitly cast every argument to its correct 64-bit C type to
        # satisfy the variadic nature of the syscall ABI.
        fd = syscall(
            ctypes.c_long(SYS_perf_event_open),
            ctypes.byref(attr),
            ctypes.c_long(0),  # pid
            ctypes.c_long(-1),  # cpu
            ctypes.c_long(group_fd),
            ctypes.c_long(0),  # flags
        )

        if fd < 0:
            return None

        return fd

    # See 'man perf_event_open' for details
    PERF_TYPE_HARDWARE = 0
    PERF_COUNT_HW_REF_CPU_CYCLES = 9
    PERF_COUNT_HW_INSTRUCTIONS = 1

    class _PerfEventAttrUnion(ctypes.Union):
        _fields_ = [
            ("sample_period", ctypes.c_uint64),
            ("sample_freq", ctypes.c_uint64),
        ]

    class PerfEventAttr(ctypes.Structure):
        # This structure must match the C structure, otherwise the kernel will
        # reject it with E2BIG, because the size field will not be what it
        # expects. spell-checker: ignore clockid,regs
        _fields_ = [
            ("type", ctypes.c_uint32),
            ("size", ctypes.c_uint32),
            ("config", ctypes.c_uint64),
            ("read_format", ctypes.c_uint64),
            ("_union1", _PerfEventAttrUnion),
            ("sample_type", ctypes.c_uint64),
            ("read_format", ctypes.c_uint64),
            ("flags", ctypes.c_uint64),
            # More fields for the union
            ("wakeup_events", ctypes.c_uint32),
            ("wakeup_watermark", ctypes.c_uint32),
            ("bp_type", ctypes.c_uint32),
            ("bp_addr", ctypes.c_uint64),
            ("bp_len", ctypes.c_uint64),
            ("branch_sample_type", ctypes.c_uint64),
            ("sample_regs_user", ctypes.c_uint64),
            ("sample_stack_user", ctypes.c_uint32),
            ("clockid", ctypes.c_int32),
            ("sample_regs_intr", ctypes.c_uint64),
            ("aux_watermark", ctypes.c_uint32),
            ("sample_max_stack", ctypes.c_uint16),
            ("__reserved_2", ctypes.c_uint16),
            ("aux_sample_size", ctypes.c_uint32),
            ("__reserved_3", ctypes.c_uint32),
            ("sig_data", ctypes.c_uint64),
        ]

    # --- Define ioctl magic numbers ---
    # These are from <linux/perf_event.h>
    PERF_EVENT_IOC_ENABLE = 0x2400
    PERF_EVENT_IOC_DISABLE = 0x2401
    PERF_EVENT_IOC_RESET = 0x2403

    class PerfCounter(object):
        """
        A simple in-process wrapper for a single perf counter.
        """

        def __init__(self, config, group_fd=-1, read_format=0):
            self.fd = getPerfFileHandle(
                config, group_fd=group_fd, read_format=read_format
            )

            assert self.fd is not None

        def reset(self):
            fcntl.ioctl(self.fd, PERF_EVENT_IOC_RESET, 0)

        def start(self):
            self.reset()
            fcntl.ioctl(self.fd, PERF_EVENT_IOC_ENABLE, 0)

        def stop(self):
            try:
                fcntl.ioctl(self.fd, PERF_EVENT_IOC_DISABLE, 0)
            except OSError:
                pass

        def read(self):
            # Reads the 8-byte (64-bit) counter value
            count_bytes = os.read(self.fd, 8)
            return struct.unpack("Q", count_bytes)[0]

        def close(self):
            os.close(self.fd)

    class PerfCounters(object):
        def __init__(self):
            # Create the instruction counter as the group leader.
            self.instr_counter = PerfCounter(config=PERF_COUNT_HW_INSTRUCTIONS)
            # Create the cycle counter as a member of the same group.
            self.cycle_counter = PerfCounter(
                config=PERF_COUNT_HW_REF_CPU_CYCLES, group_fd=self.instr_counter.fd
            )

        def start(self):
            # Enabling the group leader enables all counters in the group.
            self.instr_counter.start()

        def stop(self):
            # Disabling the group leader disables all counters in the group.
            self.instr_counter.stop()

        def getValues(self):
            instr_count = self.instr_counter.read()
            cycle_count = self.cycle_counter.read()

            self.instr_counter.close()
            self.cycle_counter.close()

            return instr_count, cycle_count


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
